/**
 * Copyright  2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.aspect.axiom.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;
import javax.xml.transform.sax.SAXSource;

import org.apache.axiom.om.impl.builder.StAXBuilder;
import org.apache.axiom.om.impl.dom.factory.OMDOMFactory;
import org.bluestemsoftware.open.eoa.aspect.axiom.DocumentElement;
import org.bluestemsoftware.open.eoa.aspect.axiom.ParsedContent;
import org.bluestemsoftware.specification.eoa.component.message.rt.Content;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Overcomes issue with axiom StAXUtils class which sets context classloader to its defining
 * classloader prior to invoking XMLInputFactory.newInstance() and/or
 * XMLOutputFactory.newInstance(), the java service provider discovery mechanism breaks.
 * Because an instance of this class is on SHARED classpath, as it should be (its defined
 * within axiom-api artifact), we would not be able to specify woodstox as our stax factory
 * impl, i.e. service provider file is on DEPLOYMENT classpath. Also axiom StAXUtils class
 * pools factories, but woodstox factories are threadsafe, i.e. no pooling required.
 */
public class STAXUtils {

    private static final String INPUT_CLASS = "com.ctc.wstx.stax.WstxInputFactory";
    private static XMLInputFactory xmlInputFactory;
    private static final String OUTPUT_CLASS = "com.ctc.wstx.stax.WstxOutputFactory";
    private static XMLOutputFactory xmlOutputFactory;
    private static final String EOF_EXCEPTION = "com.ctc.wstx.exc.WstxEOFException";

    public static XMLStreamReader createXMLStreamReader(InputStream in, String encoding) throws XMLStreamException {
        return getInputFactory().createXMLStreamReader(in, encoding);
    }
    
    public static XMLStreamWriter createXMLStreamWriter(OutputStream out, String encoding) throws XMLStreamException {
        return getOutputFactory().createXMLStreamWriter(out, encoding);
    }
    
    public static Content parseContent(OMDOMFactory factory, InputStream inputStream, String encoding) throws SAXException, IOException {

        // a one-to-one relationship exists between factory and a document. we cannot pass
        // factory to builder, i.e because getDocumentElement would return the message
        // element - not the parsed content. so ... we create a new factory and detach the
        // content from document created by builder and set owner document to document
        // retrieved from factory, i.e. the document that's parsing the content
        
        try {
            InputSource is = new InputSource(inputStream);
            is.setEncoding(encoding);
            XMLStreamReader parser = getInputFactory().createXMLStreamReader(new SAXSource(is));
            OMDOMFactory newFactory = factory.getClass().newInstance();
            StAXBuilder builder = new ContentBuilder(newFactory, parser);
            DocumentElement documentElement = (DocumentElement)builder.getDocumentElement();
            documentElement.detachFromDocument();
            return ((ParsedContent)documentElement).setOwnerDocument(factory.getDocument());
        } catch (Throwable th) {
            Throwable cause = th.getCause();
            if (cause != null && cause.getClass().getName().equals(EOF_EXCEPTION)) {
                // file is empty, i.e. empty payload. return null
                return null;
            }
            throw new IOException("Error reading message. " + th.getMessage());
        }

    }

    private static XMLInputFactory getInputFactory() throws XMLStreamException {

        // note that chosen parser implementation is woodstox. wstx parser factory is thread
        // safe and can be used concurrently, the bundled jdk streaming factory (as of 1.6.0),
        // is NOT threadsafe

        if (xmlInputFactory == null) {
            
            Class temp;
            try {
                temp = Thread.currentThread().getContextClassLoader().loadClass(INPUT_CLASS);
            } catch (ClassNotFoundException cfe) {
                throw new IllegalStateException("Deployment classloader must scope factory class "
                        + INPUT_CLASS
                        + " to deployment classpath");
            }
            try {
                xmlInputFactory = (XMLInputFactory)temp.newInstance();
            } catch (Exception ex) {
                throw new XMLStreamException("Error instantiating parser factory. " + ex);
            }
            
            xmlInputFactory.setProperty(XMLInputFactory.IS_NAMESPACE_AWARE, Boolean.TRUE);
            xmlInputFactory.setProperty(XMLInputFactory.IS_VALIDATING, Boolean.FALSE);
            xmlInputFactory.setProperty(XMLInputFactory.SUPPORT_DTD, Boolean.FALSE);
            
            //adds 10% overhead
            //xmlInputFactory.setProperty(XMLInputFactory.IS_COALESCING, Boolean.TRUE);
            
        }

        return xmlInputFactory;

    }
    
    private static XMLOutputFactory getOutputFactory() throws XMLStreamException {

        // note that chosen parser implementation is woodstox. wstx parser factory is thread
        // safe and can be used concurrently, the bundled jdk streaming factory (as of 1.6.0),
        // is NOT threadsafe

        if (xmlOutputFactory == null) {
            Class temp;
            try {
                temp = Thread.currentThread().getContextClassLoader().loadClass(OUTPUT_CLASS);
            } catch (ClassNotFoundException cfe) {
                throw new IllegalStateException("Deployment classloader must scope factory class "
                        + OUTPUT_CLASS
                        + " to deployment classpath");
            }
            try {
                xmlOutputFactory = (XMLOutputFactory)temp.newInstance();
            } catch (Exception ex) {
                throw new XMLStreamException("Error instantiating parser factory. " + ex);
            }
            //xmlOutputFactory.setProperty(XMLOutputFactory.IS_REPAIRING_NAMESPACES, Boolean.TRUE);
        }

        return xmlOutputFactory;

    }
}
